@@ -4,16 +4,16 @@ import (
 	"bufio"
 	"bytes"
 	"encoding/json"
+	"errors"
 	"fmt"
 	"io"
 	"sort"
 	"strings"
+	"sync"
 	"time"
 
-	"github.com/jonboulle/clockwork"
-	"github.com/pkg/errors"
 	"golang.org/x/sync/errgroup"
-	"gotest.tools/gotestsum/log"
+	"gotest.tools/gotestsum/internal/log"
 )
 
 // Action of TestEvent
@@ -31,6 +31,16 @@ const (
 	ActionSkip   Action = "skip"
 )
 
+// IsTerminal returns true if the Action is one of: pass, fail, skip.
+func (a Action) IsTerminal() bool {
+	switch a {
+	case ActionPass, ActionFail, ActionSkip:
+		return true
+	default:
+		return false
+	}
+}
+
 // TestEvent is a structure output by go tool test2json and go test -json.
 type TestEvent struct {
 	// Time encoded as an RFC3339-format string
@@ -44,6 +54,8 @@ type TestEvent struct {
 	Output string
 	// raw is the raw JSON bytes of the event
 	raw []byte
+	// RunID from the ScanConfig which produced this test event.
+	RunID int
 }
 
 // PackageEvent returns true if the event is a package start or end event
@@ -69,9 +81,12 @@ type Package struct {
 	Skipped []TestCase
 	Passed  []TestCase
 
+	// elapsed time reported by the pass or fail event for the package.
+	elapsed time.Duration
+
 	// mapping of root TestCase ID to all sub test IDs. Used to mitigate
-	// github.com/golang/go/issues/29755.
-	// In the future when that bug is fixed this mapping can likely be removed.
+	// github.com/golang/go/issues/29755, and github.com/golang/go/issues/40771.
+	// In the future when those bug are fixed this mapping can likely be removed.
 	subTests map[int][]int
 
 	// output printed by test cases, indexed by TestCase.ID. Package output is
@@ -86,6 +101,14 @@ type Package struct {
 	action Action
 	// cached is true if the package was marked as (cached)
 	cached bool
+	// panicked is true if the package, or one of the tests in the package,
+	// contained output that looked like a panic. This is used to mitigate
+	// github.com/golang/go/issues/45508. This field may be removed in the future
+	// if the issue is fixed in Go.
+	panicked bool
+	// shuffleSeed is the seed used to shuffle the tests. The value is set when
+	// tests are run with -shuffle
+	shuffleSeed string
 }
 
 // Result returns if the package passed, failed, or was skipped because there
@@ -94,13 +117,10 @@ func (p *Package) Result() Action {
 	return p.action
 }
 
-// Elapsed returns the sum of the elapsed time for all tests in the package.
+// Elapsed returns the elapsed time of the package, as reported by the
+// pass or fail event for the package.
 func (p *Package) Elapsed() time.Duration {
-	elapsed := time.Duration(0)
-	for _, testcase := range p.TestCases() {
-		elapsed += testcase.Elapsed
-	}
-	return elapsed
+	return p.elapsed
 }
 
 // TestCases returns all the test cases.
@@ -119,7 +139,7 @@ func (p *Package) TestCases() []TestCase {
 // provides a little more safety if that ever changes.
 func (p *Package) LastFailedByName(name string) TestCase {
 	for i := len(p.Failed) - 1; i >= 0; i-- {
-		if p.Failed[i].Test == name {
+		if p.Failed[i].Test.Name() == name {
 			return p.Failed[i]
 		}
 	}
@@ -138,16 +158,16 @@ func (p *Package) Output(id int) string {
 // As a workaround for test output being attributed to the wrong subtest, if:
 //   - the TestCase is a root TestCase (not a subtest), and
 //   - the TestCase has no subtest failures;
+//
 // then all output for every subtest under the root test is returned.
 // See https://github.com/golang/go/issues/29755.
 func (p *Package) OutputLines(tc TestCase) []string {
-	_, sub := splitTestName(tc.Test)
 	lines := p.output[tc.ID]
 
 	// If this is a subtest, or a root test case with subtest failures the
 	// subtest failure output should contain the relevant lines, so we don't need
 	// extra lines.
-	if sub != "" || tc.hasSubTestFailed {
+	if tc.Test.IsSubTest() || tc.hasSubTestFailed {
 		return lines
 	}
 
@@ -160,19 +180,33 @@ func (p *Package) OutputLines(tc TestCase) []string {
 }
 
 func (p *Package) addOutput(id int, output string) {
+	if strings.HasPrefix(output, "panic: ") {
+		p.panicked = true
+	}
 	// TODO: limit size of buffered test output
 	p.output[id] = append(p.output[id], output)
 }
 
-// splitTestName into root test name and any subtest names.
-func splitTestName(name string) (root, sub string) {
-	parts := strings.SplitN(name, "/", 2)
+type TestName string
+
+func (n TestName) Split() (root string, sub string) {
+	parts := strings.SplitN(string(n), "/", 2)
 	if len(parts) < 2 {
-		return name, ""
+		return string(n), ""
 	}
 	return parts[0], parts[1]
 }
 
+// IsSubTest returns true if the name indicates the test is a subtest run using
+// t.Run().
+func (n TestName) IsSubTest() bool {
+	return strings.Contains(string(n), "/")
+}
+
+func (n TestName) Name() string {
+	return string(n)
+}
+
 func (p *Package) removeOutput(id int) {
 	delete(p.output, id)
 
@@ -207,20 +241,53 @@ const neverFinished time.Duration = -1
 // in some cases, when a test panics.
 func (p *Package) end() []TestEvent {
 	result := make([]TestEvent, 0, len(p.running))
-	for _, tc := range p.running {
+	for k, tc := range p.running {
+		if tc.Test.IsSubTest() && rootTestPassed(p, tc) {
+			// mitigate github.com/golang/go/issues/40771 (gotestsum/issues/141)
+			// by skipping missing subtest end events when the root test passed.
+			continue
+		}
+
 		tc.Elapsed = neverFinished
 		p.Failed = append(p.Failed, tc)
 
 		result = append(result, TestEvent{
 			Action:  ActionFail,
 			Package: tc.Package,
-			Test:    tc.Test,
+			Test:    tc.Test.Name(),
 			Elapsed: float64(neverFinished),
 		})
+		delete(p.running, k)
 	}
 	return result
 }
 
+// rootTestPassed looks for the root test associated with subtest and returns
+// true if the root test passed. This is used to mitigate
+// github.com/golang/go/issues/40771 (gotestsum/issues/141) and may be removed
+// in the future since that issue was patched in go1.16.
+//
+// This function is slightly expensive because it has to scan every test in the
+// package, but it should only run in the rare case where a subtest was missing
+// an end event. Spending a little more time in that rare case is probably better
+// than keeping extra mapping of tests in all cases.
+func rootTestPassed(p *Package, subtest TestCase) bool {
+	root, _ := subtest.Test.Split()
+
+	for _, tc := range p.Passed {
+		if tc.Test.Name() != root {
+			continue
+		}
+
+		for _, subID := range p.subTests[tc.ID] {
+			if subID == subtest.ID {
+				return true
+			}
+		}
+	}
+	return false
+}
+
 // TestCase stores the name and elapsed time for a test case.
 type TestCase struct {
 	// ID is unique ID for each test case. A test run may include multiple instances
@@ -229,11 +296,15 @@ type TestCase struct {
 	// test case.
 	ID      int
 	Package string
-	Test    string
+	Test    TestName
 	Elapsed time.Duration
+	// RunID from the ScanConfig which produced this test case.
+	RunID int
 	// hasSubTestFailed is true when a subtest of this TestCase has failed. It is
 	// used to find root TestCases which have no failing subtests.
 	hasSubTestFailed bool
+	// Time when the test was run.
+	Time time.Time
 }
 
 func newPackage() *Package {
@@ -246,10 +317,12 @@ func newPackage() *Package {
 
 // Execution of one or more test packages
 type Execution struct {
-	started  time.Time
-	packages map[string]*Package
-	errors   []string
-	done     bool
+	started    time.Time
+	packages   map[string]*Package
+	errorsLock sync.RWMutex
+	errors     []string
+	done       bool
+	lastRunID  int
 }
 
 func (e *Execution) add(event TestEvent) {
@@ -259,56 +332,75 @@ func (e *Execution) add(event TestEvent) {
 		e.packages[event.Package] = pkg
 	}
 	if event.PackageEvent() {
-		e.addPackageEvent(pkg, event)
+		pkg.addEvent(event)
 		return
 	}
 	pkg.addTestEvent(event)
 }
 
-func (e *Execution) addPackageEvent(pkg *Package, event TestEvent) {
+func (p *Package) addEvent(event TestEvent) {
 	switch event.Action {
 	case ActionPass, ActionFail:
-		pkg.action = event.Action
+		p.action = event.Action
+		p.elapsed = elapsedDuration(event.Elapsed)
 	case ActionOutput:
 		if isCoverageOutput(event.Output) {
-			pkg.coverage = strings.TrimRight(event.Output, "\n")
+			p.coverage = strings.TrimRight(event.Output, "\n")
 		}
-		if isCachedOutput(event.Output) {
-			pkg.cached = true
+		if strings.Contains(event.Output, "\t(cached)") {
+			p.cached = true
 		}
-		pkg.addOutput(0, event.Output)
+		if isShuffleSeedOutput(event.Output) {
+			p.shuffleSeed = strings.TrimRight(event.Output, "\n")
+		}
+		p.addOutput(0, event.Output)
 	}
 }
 
-// nolint: gocyclo
-func (p *Package) addTestEvent(event TestEvent) {
-	tc := p.running[event.Test]
-	root, subTest := splitTestName(event.Test)
+func (p *Package) newTestCaseFromEvent(event TestEvent) TestCase {
+	// Incremental total before using it as the ID, because ID 0 is used for
+	// the package output
+	p.Total++
+	return TestCase{
+		Package: event.Package,
+		Test:    TestName(event.Test),
+		ID:      p.Total,
+		RunID:   event.RunID,
+		Time:    event.Time,
+	}
+}
 
-	switch event.Action {
-	case ActionRun:
-		// Incremental total before using it as the ID, because ID 0 is used for
-		// the package output
-		p.Total++
-		tc := TestCase{
-			Package: event.Package,
-			Test:    event.Test,
-			ID:      p.Total,
-		}
+func (p *Package) addTestEvent(event TestEvent) {
+	if event.Action == ActionRun {
+		tc := p.newTestCaseFromEvent(event)
 		p.running[event.Test] = tc
 
-		if subTest != "" {
+		if tc.Test.IsSubTest() {
+			root, _ := TestName(event.Test).Split()
 			rootID := p.running[root].ID
 			p.subTests[rootID] = append(p.subTests[rootID], tc.ID)
 		}
 		return
+	}
+
+	tc := p.running[event.Test]
+	// This appears to be a bug in 'go test' or test2json. This test is missing
+	// an Action=run event. Create one on the first event received from the test.
+	if tc.ID == 0 {
+		tc = p.newTestCaseFromEvent(event)
+		p.running[event.Test] = tc
+	}
+
+	switch event.Action {
 	case ActionOutput, ActionBench:
+		tc := p.running[event.Test]
 		p.addOutput(tc.ID, event.Output)
 		return
 	case ActionPause, ActionCont:
 		return
 	}
 
+	// the event.Action must be one of the three "test end" events
 	delete(p.running, event.Test)
 	tc.Elapsed = elapsedDuration(event.Elapsed)
 
@@ -317,7 +409,8 @@ func (p *Package) addTestEvent(event TestEvent) {
 		p.Failed = append(p.Failed, tc)
 
 		// If this is a subtest, mark the root test as having a failed subtest
-		if subTest != "" {
+		if tc.Test.IsSubTest() {
+			root, _ := TestName(event.Test).Split()
 			rootTestCase := p.running[root]
 			rootTestCase.hasSubTestFailed = true
 			p.running[root] = rootTestCase
@@ -331,14 +424,12 @@ func (p *Package) addTestEvent(event TestEvent) {
 		// Do not immediately remove output for subtests, to work around a bug
 		// in 'go test' where output is attributed to the wrong sub test.
 		// github.com/golang/go/issues/29755.
-		if subTest != "" {
+		if tc.Test.IsSubTest() {
 			return
 		}
 
 		// Remove test output once a test passes, it wont be used.
 		p.removeOutput(tc.ID)
-		// Remove subtest mapping, it is only used when a test fails.
-		delete(p.subTests, tc.ID)
 	}
 }
 
@@ -349,11 +440,15 @@ func elapsedDuration(elapsed float64) time.Duration {
 func isCoverageOutput(output string) bool {
 	return all(
 		strings.HasPrefix(output, "coverage:"),
-		strings.HasSuffix(output, "% of statements\n"))
+		strings.Contains(output, "% of statements"))
 }
 
-func isCachedOutput(output string) bool {
-	return strings.Contains(output, "\t(cached)")
+func isShuffleSeedOutput(output string) bool {
+	return strings.HasPrefix(output, "-test.shuffle ")
+}
+
+func isWarningNoTestsToRunOutput(output string) bool {
+	return output == "testing: warning: no tests to run\n"
 }
 
 // OutputLines returns the full test output for a test as an slice of lines.
@@ -375,11 +470,11 @@ func (e *Execution) Packages() []string {
 	return sortedKeys(e.packages)
 }
 
-var clock = clockwork.NewRealClock()
+var timeNow = time.Now
 
 // Elapsed returns the time elapsed since the execution started.
 func (e *Execution) Elapsed() time.Duration {
-	return clock.Now().Sub(e.started)
+	return timeNow().Sub(e.started)
 }
 
 // Failed returns a list of all the failed test cases.
@@ -400,6 +495,18 @@ func (e *Execution) Failed() []TestCase {
 	return failed
 }
 
+// FilterFailedUnique filters a slice of failed TestCases by removing root test
+// case that have failed subtests.
+func FilterFailedUnique(tcs []TestCase) []TestCase {
+	var result []TestCase
+	for _, tc := range tcs {
+		if !tc.hasSubTestFailed {
+			result = append(result, tc)
+		}
+	}
+	return result
+}
+
 func sortedKeys(pkgs map[string]*Package) []string {
 	keys := make([]string, 0, len(pkgs))
 	for key := range pkgs {
@@ -432,14 +539,29 @@ func (e *Execution) addError(err string) {
 	if strings.HasPrefix(err, "# ") {
 		return
 	}
+	e.errorsLock.Lock()
 	e.errors = append(e.errors, err)
+	e.errorsLock.Unlock()
 }
 
 // Errors returns a list of all the errors.
 func (e *Execution) Errors() []string {
+	e.errorsLock.RLock()
+	defer e.errorsLock.RUnlock()
 	return e.errors
 }
 
+// HasPanic returns true if at least one package had output that looked like a
+// panic.
+func (e *Execution) HasPanic() bool {
+	for _, pkg := range e.packages {
+		if pkg.panicked {
+			return true
+		}
+	}
+	return false
+}
+
 func (e *Execution) end() []TestEvent {
 	e.done = true
 	var result []TestEvent // nolint: prealloc
@@ -449,17 +571,24 @@ func (e *Execution) end() []TestEvent {
 	return result
 }
 
+func (e *Execution) Started() time.Time {
+	return e.started
+}
+
 // newExecution returns a new Execution and records the current time as the
 // time the test execution started.
 func newExecution() *Execution {
 	return &Execution{
-		started:  clock.Now(),
+		started:  timeNow(),
 		packages: make(map[string]*Package),
 	}
 }
 
 // ScanConfig used by ScanTestOutput.
 type ScanConfig struct {
+	// RunID is a unique identifier for the run. It may be set to the pid of the
+	// process, or some other identifier. It will stored as the TestCase.RunID.
+	RunID int
 	// Stdout is a reader that yields the test2json output stream.
 	Stdout io.Reader
 	// Stderr is a reader that yields stderr from the 'go test' process. Often
@@ -470,6 +599,11 @@ type ScanConfig struct {
 	// Execution to populate while scanning. If nil a new one will be created
 	// and returned from ScanTestOutput.
 	Execution *Execution
+	// Stop is called when ScanTestOutput fails during a scan.
+	Stop func()
+	// IgnoreNonJSONOutputLines causes ScanTestOutput to ignore non-JSON lines received from
+	// the Stdout reader. Instead of causing an error, the lines will be sent to Handler.Err.
+	IgnoreNonJSONOutputLines bool
 }
 
 // EventHandler is called by ScanTestOutput for each event and write to stderr.
@@ -496,27 +630,39 @@ func ScanTestOutput(config ScanConfig) (*Execution, error) {
 	if config.Stderr == nil {
 		config.Stderr = new(bytes.Reader)
 	}
+	if config.Stop == nil {
+		config.Stop = func() {}
+	}
 	execution := config.Execution
 	if execution == nil {
 		execution = newExecution()
 	}
+	execution.done = false
+	execution.lastRunID = config.RunID
+
 	var group errgroup.Group
 	group.Go(func() error {
-		return readStdout(config, execution)
+		return stopOnError(config.Stop, readStdout(config, execution))
 	})
 	group.Go(func() error {
-		return readStderr(config, execution)
+		return stopOnError(config.Stop, readStderr(config, execution))
 	})
-	if err := group.Wait(); err != nil {
-		return execution, err
-	}
+
+	err := group.Wait()
 	for _, event := range execution.end() {
 		if err := config.Handler.Event(event, execution); err != nil {
 			return execution, err
 		}
 	}
+	return execution, err
+}
 
-	return execution, nil
+func stopOnError(stop func(), err error) error {
+	if err != nil {
+		stop()
+		return err
+	}
+	return nil
 }
 
 func readStdout(config ScanConfig, execution *Execution) error {
@@ -530,15 +676,24 @@ func readStdout(config ScanConfig, execution *Execution) error {
 			config.Handler.Err(errBadEvent.Error() + ": " + scanner.Text())
 			continue
 		case err != nil:
-			return errors.Wrapf(err, "failed to parse test output: %s", string(raw))
+			if config.IgnoreNonJSONOutputLines {
+				// nolint: errcheck
+				config.Handler.Err(string(raw))
+				continue
+			}
+			return fmt.Errorf("failed to parse test output: %s: %w", string(raw), err)
 		}
 
+		event.RunID = config.RunID
 		execution.add(event)
 		if err := config.Handler.Event(event, execution); err != nil {
 			return err
 		}
 	}
-	return errors.Wrap(scanner.Err(), "failed to scan test output")
+	if err := scanner.Err(); err != nil {
+		return fmt.Errorf("failed to scan test output: %w", err)
+	}
+	return nil
 }
 
 func readStderr(config ScanConfig, execution *Execution) error {
@@ -551,6 +706,9 @@ func readStderr(config ScanConfig, execution *Execution) error {
 		if isGoModuleOutput(line) {
 			continue
 		}
+		if strings.HasPrefix(line, "warning:") {
+			continue
+		}
 		execution.addError(line)
 	}
 	if err := scanner.Err(); err != nil {
